////////////////////////////////////////////////////////////////////// 
// KinshipEmp.cpp 
// (c) 2012-2013 Shuang Feng, Dajiang Liu, Goncalo Abecasis
// 
// This file is distributed as part of the RareMetalWorker source code package   
// and may not be redistributed in any form, without prior written    

// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile RareMetalWorker.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Wednesday November 28, 2012
// 

#include "KinshipEmp.h"
#include "MathMatrix.h"
#include "PedigreeAlleleFreq.h"
#include "OutputKin.h"
#include "PreMeta.h"
#include <string>
#include <cstring>
#include <iostream>
#include <stdlib.h>
#include <vector>
#include "math.h"
#include <algorithm>
#include "InputFile.h"
#include "StringHash.h"

double KinshipEmp::q = 0.05;
double KinshipEmp::miss = 0.05;

void KinshipEmp::ReadEmpKin(FILE * log)
{
   StringArray tmp;

   //Get the pids from kinship matrix
   IFILE empIn;
   empIn = ifopen(FastTransform::readInEmp.c_str(),"r");

   if(empIn==NULL)
   {
      fprintf(log,"FATAL ERROR! Empirical kinship can not be obtained from file: %s.\n",FastTransform::readInEmp.c_str());
      error("Empirical kinship can not be obtained from file: %s.\n",FastTransform::readInEmp.c_str());
   }
   StringArray IDs;
   while (!ifeof(empIn))
   {
      String buffer;
      buffer.ReadLine(empIn);
      tmp.Clear();
      tmp.AddTokens(buffer, SEPARATORS);
      if(tmp.Length()>0)
      {
	 for(int i=0;i<tmp.Length();i++)
	 {
	    IDs.Push(tmp[i]);
	    IDFromEmp.SetInteger(tmp[i],i);
	 }
	 break;
      }
   }
   ifclose(empIn);

   //Fill in the empirical kinship matrix
   allPairs.Dimension(tmp.Length(),tmp.Length(),0.0);

   empIn = ifopen(FastTransform::readInEmp.c_str(),"r");
   int idx=0;
   while (!ifeof(empIn))
   {
      idx++;
      String buffer;
      buffer.ReadLine(empIn);
      if(idx==1)
	 continue;
      if(buffer.Find("nan")!=-1)
      {
error("Empirical kinship is not all numerical. Please check to make sure your file does not contain \"nan\"s.\n");
      }
      tmp.Clear();
      tmp.AddTokens(buffer, SEPARATORS);
      for(int i=0;i<tmp.Length();i++)
	 allPairs[idx-2][i] = tmp[i].AsDouble();
   }
   ifclose(empIn);

   if(AutoFit::fitX)
   {
      String readInEmpX;
      readInEmpX = FastTransform::readInEmp + "X";
      empIn = ifopen(readInEmpX,"r");
      if(empIn)
      {
	 fprintf(log,"Cannot find file %s for kinship X. Please check file name.\n",readInEmpX.c_str());
	 error("Error! Cannot find file %s for kinship X. Please check file name.\n",readInEmpX.c_str());
      }

      while (!ifeof(empIn))
      {
	 String buffer;
	 buffer.ReadLine(empIn);
	 tmp.Clear();
	 tmp.AddTokens(buffer, SEPARATORS);
	 if(tmp.Length()>0)
	 {
	    for(int i=0;i<tmp.Length();i++)
	       IDFromEmpX.SetInteger(tmp[i],i);
	    break;
	 }
      }
      ifclose(empIn);

      //Fill in the empirical kinship matrix
      allPairsX.Dimension(tmp.Length(),tmp.Length());

      empIn = ifopen(readInEmpX,"r");
      idx=0;
      while (!ifeof(empIn))
      {
	 idx++;
	 String buffer;
	 buffer.ReadLine(empIn);
	 if(idx==1)
	    continue;
	 tmp.Clear();
	 tmp.AddTokens(buffer, SEPARATORS);
	 for(int i=0; i<tmp.Length(); i++)
	 {
	    allPairsX[idx-2][i] = tmp[i].AsDouble();
	 }
      }
      ifclose(empIn);
   }
}

void KinshipEmp::SetupEmpKin(Pedigree & ped, IntArray & genotypedSamplePED, IntArray & genotypedSampleVCF, StringIntHash & samplePEDIDHash, StringIntHash & skippedSNPs,FILE * log)
{
   if(FastTransform::readInEmp != "")
   {
      ReadEmpKin(log);
   }
   else
   {
      if(PreMeta::genoFromPed)
      {
	 if(AutoFit::fitX)
	    SetupPEDX(ped,genotypedSamplePED,log);
	 SetupPEDAuto(ped,genotypedSamplePED,log);
      }
      else
      {
	 if(AutoFit::fitX)
	    SetupVCFX(ped,genotypedSampleVCF,samplePEDIDHash,skippedSNPs,log);
	 SetupVCFAuto(ped,genotypedSampleVCF,log);
      }
   }
}

void KinshipEmp::SetupPEDAuto(Pedigree & ped,IntArray & genotypedSamplePED, FILE * log)
{
   printf("Calculating empirical kinship matrix ...");
   fprintf(log,"Calculating empirical kinship matrix ...");

   int N=0;
   int total_n = genotypedSamplePED.Length();

   allPairs.Dimension(total_n, total_n);
   allPairs.Zero();
   Vector X;
   X.Dimension(total_n);

   String chr;
   int pos;

   for(int i=0;i<ped.markerCount;i++)
   {
      StringArray markerName;
      markerName.AddTokens(ped.markerNames[i],":");

      if(markerName.Length()<=1)
      {
	 printf("WARNING: marker name %s in DAT file is not in CHR:POS format and will be skipped from calculating empirical kinsihp.\n",ped.markerNames[i].c_str());
	 fprintf(log,"WARNING: marker name %s in DAT file is not in CHR:POS format and will be skipped from calculating empirical kinship.\n",ped.markerNames[i].c_str());
	 continue;
      }

      chr = markerName[0];
      pos = atoi(markerName[1].c_str());

      if(chr==PreMeta::xLabel && pos>=PreMeta::Xstart && pos<=PreMeta::Xend)
	 continue;

      //X.Clear();
      bool missingStatus=false;
      int alleleCount = CountAlleles(ped, i);
      //if no one is genotyped, mornomorphic, or marker is not biallelic
      if(alleleCount!=2)
	 continue;

      double mean=0.0;
      double variance = 0.0;
      double mean_sqr = 0.0;

      //now fill in the genotypes
      int n_=0,nmiss=0;
      double maf = 0.0;
      //only generate matrix for genotyped individuals
      for(int p=0;p<total_n;p++)
      {
	 if(!ped[genotypedSamplePED[p]].isGenotyped(i))
	 {
	    X[p] = -1.0;
	    nmiss++;
	    if(nmiss>total_n*miss)
	    {
	       missingStatus=true;
	       break;
	    }
	 }
	 else if(ped[genotypedSamplePED[p]].markers[i].isHeterozygous())
	 {
	    X[p]=1.0;
	    mean += X[p];
	    mean_sqr += X[p]*X[p];
	    n_++;
	 }
	 else if(ped[genotypedSamplePED[p]].markers[i].isHomozygous())
	 {
	    if(ped[genotypedSamplePED[p]].markers[i].hasAllele(2))
	    {
	       X[p]=2.0;
	       mean += X[p];
	       mean_sqr += X[p]*X[p];
	       n_++;
	    }
	    else if(ped[genotypedSamplePED[p]].markers[i].hasAllele(1))
	    {
	       X[p]=0.0;
	       n_++;
	    }
	 }
	 else
	 {
	    X[p]=0.0;
	    n_++;
	 }
      }

      if(missingStatus)
	 continue;

      maf = mean/(2.0*n_);

      if(nmiss>total_n*miss || maf<q || 1.0-maf<q)
	 continue;

      N++;
      mean /= n_;
      variance = mean_sqr/n_ - mean*mean;

      for(int p=0;p<total_n;p++)
      {
	 if(X[p]==-1.0)
	    X[p] =0.0;
	 else
	    X[p] -= mean;
      }

      for(int r=0;r<total_n;r++)
      {
	 for(int c=r;c<total_n;c++)
	 {
	    allPairs[r][c] += X[r]*X[c]/variance;
	 }
      }
   }

   if(N==0)
      error("ERROR! No qualified marker was included in calculating empirical kinsihp matrix.\n");

   printf("  %d markers were included to calculate empirical kinship.\n",N);
   fprintf(log,"  %d markers were included to calculate empirical kinship.\n",N);

   for(int i=0;i<total_n;i++)
      for(int j=i+1;j<total_n;j++)
	 allPairs[j][i] = allPairs[i][j];

   allPairs.Multiply(1.0/N);

   printf("completed. \n ");
   fprintf(log,"completed. \n ");
   printf("  %d markers were included to calculate empirical kinship.\n",N);
   fprintf(log,"  %d markers were included to calculate empirical kinship.\n",N);

   if(OutputKin::outputKin)
   {
      IFILE file = NULL;
      String filename;
      if(PreMeta::outputFile != "")
      {
	 if(PreMeta::outputFile.FastFindLastChar('.')!=-1 || PreMeta::outputFile.FastFindLastChar('/')!=-1)
	    filename.printf("%sEmpirical.Kinship.gz",(const char *) PreMeta::outputFile);
	 else
	    filename.printf("%s.Empirical.Kinship.gz",(const char *) PreMeta::outputFile);
      }
      else
      {
	 filename.printf("Empirical.Kinship.gz");
      }
      file = ifopen(filename, "wt",InputFile::GZIP);
      for(int p=0;p<total_n;p++)
      {
	 ifprintf(file,"%s ",ped[genotypedSamplePED[p]].pid.c_str());
      }
      ifprintf(file,"\n");

      for(int r=0;r<allPairs.rows;r++)
      {
	 for(int c=0;c<allPairs.cols;c++)
	 {
	    ifprintf(file,"%g ",allPairs[r][c]);
	 }
	 ifprintf(file,"\n");
      }
      ifclose(file);
      printf("  Empirical kinship saved in %s.\n",filename.c_str());
      fprintf(log,"  Empirical kinship saved in %s.\n",filename.c_str());
   }
   printf("\n");
   fprintf(log,"\n");
}

void KinshipEmp::SetupVCFAuto(Pedigree & ped, IntArray & genotypedSampleVCF,FILE * log)
{
   printf("Calculating empirical kinship matrix ...");
   fprintf(log,"Calculating empirical kinship matrix ...");
   VcfFileReader reader;
   VcfHeader header;
   VcfRecord record;
   reader.open(PreMeta::vcfInput,header);

   int N=0;
   int total_n = genotypedSampleVCF.Length();
   allPairs.Dimension(total_n,total_n);
   allPairs.Zero();

   Vector genotype;
   genotype.Dimension(total_n);

   String chr;
   int pos;

   if(PreMeta::dosage)
   {
      while(reader.readRecord(record))
      {
	 chr = record.getChromStr();
	 pos = record.get1BasedPosition();

	 if(chr==PreMeta::xLabel && pos>=PreMeta::Xstart && pos<=PreMeta::Xend)
	    continue;

	 double mean = 0.0, mean_sqr= 0.0,maf=0.0;
	 int n_=0, nmiss=0;
	 VcfRecordGenotype & genoInfo = record.getGenotypeInfo();
	 for (int p = 0; p < genotypedSampleVCF.Length(); p++)
	 {
	    int s = genotypedSampleVCF[p];
	    //fill in genotype vector for this sample
	    const std::string * geno = genoInfo.getString("DS",s);
	    if(*geno == ".")
	    {
	       genotype[p] = _NAN_;
	       nmiss++;
	    }
	    else
	    {
	       genotype[p] = atof((*geno).c_str());
	       mean += genotype[p];
	       n_++;
	       mean_sqr += genotype[p]*genotype[p];
	    }
	 }
	 maf = mean/(2.0*n_);
	 if(nmiss>total_n*miss || maf<q || 1.0-maf<q)
	    continue;

	 N++;
	 mean = maf*2.0;
	 double variance = mean_sqr/n_ - mean*mean;
	 //standardize genotype vector
	 for(int i=0;i<genotype.Length();i++)
	 {
	    if(genotype[i]==_NAN_)
	       //make sure if genotype is missing, then it is not contributing
	       genotype[i]=0.0;
	    else
	       genotype[i] -= mean;
	 }
	 //add up to kinship
	 for(int r=0;r<total_n;r++)
	 {
	    for(int c=r;c<total_n;c++)
	    {
	       allPairs[r][c] += genotype[r]*genotype[c]/variance;
	    }
	 }
      }
      for(int i=0;i<total_n;i++)
	 for(int j=i+1;j<total_n;j++)
	    allPairs[j][i] = allPairs[i][j];

      allPairs.Multiply(1.0/N);
   }
   else
   {
      IntArray counts;
      counts.Dimension(2);
      while(reader.readRecord(record))
      {
	 chr = record.getChromStr();
	 pos = record.get1BasedPosition();

	 if(chr==PreMeta::xLabel && pos>=PreMeta::Xstart && pos<=PreMeta::Xend)
	    continue;

	 double mean = 0.0, mean_sqr= 0.0,maf=0.0;
	 int n_=0, nmiss=0;
	 counts.Zero();
	 genotype.Zero();

	 bool skipSNP=false;
	 for(int s=0;s<total_n;s++)
	 {
	    bool skip=false;
	    int i= genotypedSampleVCF[s];
	    int numGTs = record.getNumGTs(i);
	    for(int j = 0; j < numGTs; j++)
	    {
	       int a = record.getGT(i,j);
	       if(a==VcfGenotypeSample::MISSING_GT)
	       {
		  nmiss++;
		  if(nmiss>total_n*miss)
		  {
		     skipSNP=true;
		     break;
		  }
		  skip =true;
		  genotype[s]=_NAN_;
		  break;
	       }
	       //if marker is not biallelic, skip this marker
	       if(a>1)
	       {
		  skipSNP=true;
		  break;
	       }
	       genotype[s] += a;
	       counts[a]++;
	    }
	    if(skipSNP)
	       break;
	    if(skip)
	       continue;
	    mean+=genotype[s];
	    n_++;
	    mean_sqr += genotype[s]*genotype[s];
	 }

	 if(skipSNP)
	    continue;

	 maf = (double)counts[1]/(counts[0]+counts[1]);

	 if(nmiss > total_n*miss || maf<q || 1.0-maf<q)
	    continue;

	 N++;
	 mean /= n_;
	 double variance = mean_sqr/n_ - mean*mean;

	 for(int i=0;i<genotype.Length();i++)
	 {
	    if(genotype[i]==_NAN_)
	       genotype[i]=0.0;
	    else
	       genotype[i] -= mean;
	 }

	 for(int r=0;r<total_n;r++)
	 {
	    for(int c=r;c<total_n;c++)
	    {
	       allPairs[r][c] += genotype[r]*genotype[c]/variance;
	    }
	 }
      }

      for(int i=0;i<total_n;i++)
	 for(int j=i+1;j<total_n;j++)
	    allPairs[j][i] = allPairs[i][j];

      allPairs.Multiply(1.0/N);
   }

   if(N==0)
      error("ERROR! No qualified marker was included in calculating empirical kinsihp matrix.\n");


   printf("completed.\n  %d markers were included to calculate empirical kinship.\n",N);
   fprintf(log,"completed.\n  %d markers were included to calculate empirical kinship.\n",N);

   if(OutputKin::outputKin)
   {
      IFILE file = NULL;
      String filename;
      if(PreMeta::outputFile != "")
      {
	 if(PreMeta::outputFile.FastFindLastChar('.')!=-1 || PreMeta::outputFile.FastFindLastChar('/')!=-1)
	    filename.printf("%sEmpirical.Kinship.gz",(const char *) PreMeta::outputFile);
	 else
	    filename.printf("%s.Empirical.Kinship.gz",(const char *) PreMeta::outputFile);
      }
      else
	 filename.printf("Empirical.Kinship.gz");
      file = ifopen(filename, "wt",InputFile::GZIP);
      for (int i = 0; i < total_n; i++)
      {
	 const char * sample = header.getSampleName(genotypedSampleVCF[i]);
	 ifprintf(file,"%s ",sample);
      }
      ifprintf(file,"\n");
      for(int r=0;r<allPairs.rows;r++)
      {
	 for(int c=0;c<allPairs.cols;c++)
	 {
	    ifprintf(file,"%g ",allPairs[r][c]);
	 }
	 ifprintf(file,"\n");
      }
      ifclose(file);
      printf("  Empirical kinship saved in %s.\n",filename.c_str());
      fprintf(log,"  Empirical kinship saved in %s.\n",filename.c_str());
   }
   printf("\n");
   fprintf(log,"\n");
}

//this function calculates relationship matrix from autosomal markers
void KinshipEmp::SetupPEDX(Pedigree & ped, IntArray & genotypedSamplePED, FILE * log)
{
   printf("Calculating empirical kinshipX matrix ...");
   fprintf(log,"Calculating empirical kinshipX matrix ...");
   int N=0;
   int total_n = genotypedSamplePED.Length();

   allPairsX.Dimension(total_n, total_n);
   allPairsX.Zero();
   Vector X;
   X.Dimension(total_n);

   String chr;
   int pos;

   for(int i=0;i<ped.markerCount;i++)
   {
      StringArray markerName;
      markerName.AddTokens(ped.markerNames[i],":");

      bool maleGenotypedWrong=false;

      if(markerName.Length()<=1)
      {
	 printf("WARNING: marker name %s in DAT file is not in CHR:POS format and will be skipped from calculating empirical kinsihp.\n",ped.markerNames[i].c_str());
	 fprintf(log,"WARNING: marker name %s in DAT file is not in CHR:POS format and will be skipped from calculating empirical kinship.\n",ped.markerNames[i].c_str());
	 continue;
      }

      chr = markerName[0];
      pos = atoi(markerName[1].c_str());

      if(chr!=PreMeta::xLabel)
	 continue;

      X.Clear();
      bool missingStatus=false;
      int alleleCount = CountAlleles(ped, i);
      //if no one is genotyped, mornomorphic, or marker is not biallelic
      if(alleleCount!=2)
	 continue;

      double mean=0.0;
      double variance = 0.0;
      double mean_sqr = 0.0;

      //now fill in the genotypes
      int n_=0,nmiss=0;
      double maf = 0.0;
      //only generate matrix for genotyped individuals
      if(pos>=PreMeta::Xstart && pos<=PreMeta::Xend)
      {
	 for(int p=0;p<total_n;p++)
	 {
	    //check if this sample is male and heterozygous then skip this marker
	    if(ped[genotypedSamplePED[p]].sex==PreMeta::maleLabel && ped[genotypedSamplePED[p]].markers[i].isHeterozygous())
	    {
	       maleGenotypedWrong=true;
	       break;
	    }
	 }
      }

      if(maleGenotypedWrong)
      {
	 fprintf(log,"marker %s:%d is excluded from calculating reltionship matrix due observation of heterozygous X chromosome in at least one male sample.\n",chr.c_str(),pos);
	 continue;
      }

      for(int p=0;p<total_n;p++)
      {
	 if(!ped[genotypedSamplePED[p]].isGenotyped(i))
	 {
	    X[p] = -1.0;
	    nmiss++;
	    if(nmiss>total_n*miss)
	    {
	       missingStatus=true;
	       break;
	    }
	 }
	 else if(ped[genotypedSamplePED[p]].markers[i].isHeterozygous())
	 {
	    X[p]=1.0;
	    mean += X[p];
	    mean_sqr += X[p]*X[p];
	    n_++;
	 }
	 else if(ped[genotypedSamplePED[p]].markers[i].isHomozygous())
	 {
	    if(ped[genotypedSamplePED[p]].markers[i].hasAllele(2))
	    {
	       X[p]=2.0;
	       mean += X[p];
	       mean_sqr += X[p]*X[p];
	       n_++;
	    }
	    else if(ped[genotypedSamplePED[p]].markers[i].hasAllele(1))
	    {
	       X[p]=0.0;
	       n_++;
	    }
	 }
	 else
	 {
	    X[p]=0.0;
	    n_++;
	 }
      }

      if(missingStatus)
      {
	 fprintf(log,"marker %s:%d is excluded from calculating reltionship matrix due to missing rate > %g.\n",chr.c_str(),pos,miss);
	 continue;
      }
      maf = mean/(2.0*n_);

      if(nmiss>total_n*miss || maf<q || 1.0-maf<q)
	 continue;

      N++;
      mean /= n_;
      variance = mean_sqr/n_ - mean*mean;

      for(int p=0;p<total_n;p++)
      {
	 if(X[p]==-1.0)
	    X[p] =0.0;
	 else
	    X[p] -= mean;
      }

      for(int r=0;r<total_n;r++)
      {
	 for(int c=r;c<total_n;c++)
	 {
	    allPairsX[r][c] += X[r]*X[c]/variance;
	 }
      }
   }

   if(N==0)
      error("ERROR! No qualified marker was included in calculating empirical kinsihp matrix of X chromosome.\n");

   for(int i=0;i<total_n;i++)
      for(int j=i+1;j<total_n;j++)
	 allPairsX[j][i] = allPairsX[i][j];

   allPairsX.Multiply(1.0/N);

   printf("completed\n ");
   fprintf(log,"completed\n ");

   printf("  %d markers were included to calculate empirical kinship of X chromosome.\n",N);
   fprintf(log,"  %d markers were included to calculate empirical kinship of X chromosome.\n",N);
   if(OutputKin::outputKin)
   {
      IFILE file = NULL;
      String filename;
      if(PreMeta::outputFile != "")
      {
	 if(PreMeta::outputFile.FastFindLastChar('.')!=-1 || PreMeta::outputFile.FastFindLastChar('/')!=-1)
	    filename.printf("%sEmpirical.KinshipX.gz",(const char *) PreMeta::outputFile);
	 else
	    filename.printf("%s.Empirical.KinshipX.gz",(const char *) PreMeta::outputFile);
      }
      else
      {
	 filename.printf("Empirical.KinshipX.gz");
      }

      file = ifopen(filename, "wt", InputFile::GZIP);
      for(int p=0;p<total_n;p++)
      {
	 ifprintf(file,"%s ",ped[genotypedSamplePED[p]].pid.c_str());
      }
      ifprintf(file,"\n");

      for(int r=0;r<allPairsX.rows;r++)
      {
	 for(int c=0;c<allPairsX.cols;c++)
	 {
	    ifprintf(file,"%g ",allPairsX[r][c]);
	 }
	 ifprintf(file,"\n");
      }
      ifclose(file);
      printf("  Empirical kinshipX saved in %s.\n",filename.c_str());
      fprintf(log,"  Empirical kinshipX saved in %s.\n",filename.c_str());
   }
   printf("\n");
   fprintf(log,"\n");
}

void KinshipEmp::SetupVCFX(Pedigree & ped, IntArray & genotypedSampleVCF, StringIntHash & samplePEDIDHash, StringIntHash & skippedSNPs,FILE * log)
{
   printf("Calculating empirical kinshipX matrix ...");
   fprintf(log,"Calculating empirical kinshipX matrix ...");
   VcfFileReader reader;
   VcfHeader header;
   VcfRecord record;
   reader.open(PreMeta::vcfInput,header);
   reader.readVcfIndex();
   reader.setReadSection(PreMeta::xLabel.c_str());

   int N=0;
   int total_n = genotypedSampleVCF.Length();
   allPairsX.Dimension(total_n,total_n);
   allPairsX.Zero();

   Vector genotype;
   genotype.Dimension(total_n);

   String chr;
   int pos;

   IntArray counts;
   counts.Dimension(2);

   if(PreMeta::dosage)
   {
      while(reader.readRecord(record))
      {
	 chr = record.getChromStr();
	 pos = record.get1BasedPosition();

	 genotype.Clear();

	 if(chr!=PreMeta::xLabel)
	    continue;
	 if(pos<PreMeta::Xstart || pos>PreMeta::Xend)
	    continue;
	 double mean = 0.0, mean_sqr= 0.0,maf=0.0;
	 int n_=0, nmiss=0;
	 VcfRecordGenotype & genoInfo = record.getGenotypeInfo();
	 for (int p = 0; p < total_n; p++)
	 {   
	    int s = genotypedSampleVCF[p];

	    //fill in genotype vector for this sample
	    const std::string * geno = genoInfo.getString("DS",s);
	    if(*geno == ".")
	    {
	       genotype[p] = _NAN_;
	       nmiss++;
	    }
	    else
	    {
	       genotype[p] = atof((*geno).c_str());
	       mean += genotype[p];
	       n_++;
	       mean_sqr += genotype[p]*genotype[p];
	    }
	 }
	 maf = mean/(2.0*n_);
	 if(nmiss>total_n*miss || maf<q || 1.0-maf<q)
	    continue;
	 N++;
	 mean = maf*2.0;
	 double variance = mean_sqr/n_ - mean*mean;
	 //standardize genotype vector
	 for(int i=0;i<genotype.Length();i++)
	 {
	    if(genotype[i]==_NAN_)
	       //make sure if genotype is missing, then it is not contributing
	       genotype[i]=0.0;
	    else
	       genotype[i] -= mean;
	 }
	 //add up to kinship
	 for(int r=0;r<total_n;r++)
	 {
	    for(int c=r;c<total_n;c++)
	    {
	       allPairsX[r][c] += genotype[r]*genotype[c]/variance;
	    }
	 }
      }
   }
   else
   {
      while(reader.readRecord(record))
      {
	 chr = record.getChromStr();
	 if(chr!=PreMeta::xLabel)
	    continue;

	 pos = record.get1BasedPosition();
	 String SNPname = chr + pos;
	 if(pos>=PreMeta::Xstart && pos<=PreMeta::Xend)
	    if(skippedSNPs.Integer(SNPname)>0)
	       continue;

	 double mean = 0.0, mean_sqr= 0.0,maf=0.0;
	 int n_=0, nmiss=0;
	 counts.Zero();
	 genotype.Zero();

	 bool skipSNP=false;
	 for(int s=0;s<genotypedSampleVCF.Length();s++)
	 {
	    bool skip=false;
	    int i= genotypedSampleVCF[s];
	    int numGTs = record.getNumGTs(i);
	    //check if male are genotyped correctly

	    const char * sample = header.getSampleName(i);
	    int ped_idx = samplePEDIDHash.Integer(sample);

	    if(ped[ped_idx].sex==PreMeta::maleLabel) 
	    {
	       IntArray geno_tmp;
	       for(int j = 0; j < numGTs; j++)
	       {
		  //if marker is not biallelic, skip this marker
		  int a = record.getGT(i,j);
		  if(a>1)
		  {
		     skipSNP=true;
		     break;
		  }
		  geno_tmp.Push(a);
	       }

	       if(skipSNP)
		  break;

	       if(geno_tmp[0]+geno_tmp[1]==1)
	       {
		  skipSNP=true;
		  break;
	       }

	       if(geno_tmp[0]==VcfGenotypeSample::MISSING_GT
		     && geno_tmp[1]==VcfGenotypeSample::MISSING_GT)
	       {
		  nmiss++;
		  if(nmiss>total_n*miss)
		  {
		     skipSNP=true;
		     break;
		  }
		  skip =true;
		  genotype[s]=_NAN_;
	       }
	       else if(geno_tmp[0]==VcfGenotypeSample::MISSING_GT)
	       {
		  genotype[s]=geno_tmp[1]*2;
		  counts[1]+=2;
	       }
	       else if(geno_tmp[1]==VcfGenotypeSample::MISSING_GT)
	       {
		  genotype[s]=geno_tmp[0]*2;
		  counts[0]+=2;
	       }
	       else 
	       {
		  genotype[s] = geno_tmp[0]+geno_tmp[1];
		  counts[geno_tmp[0]]++;
		  counts[geno_tmp[1]]++;
	       }
	    }
	    else
	    {
	       for(int j = 0; j < numGTs; j++)
	       {  
		  int a = record.getGT(i,j);
		  if(a==VcfGenotypeSample::MISSING_GT)
		  {  
		     nmiss++;
		     if(nmiss>total_n*miss)
		     {  
			skipSNP=true;
			break;
		     }  
		     skip =true;
		     genotype[s]=_NAN_;
		     break;
		  }  
		  //if marker is not biallelic, skip this marker
		  if(a>1)
		  {
		     skipSNP=true;
		     break;
		  }
		  genotype[s] += a;
		  counts[a]++;
	       }
	    }
	    if(skipSNP)
	       break;
	    if(skip)
	       continue;
	    mean+=genotype[s];
	    n_++;
	    mean_sqr += genotype[s]*genotype[s];
	 }

	 if(skipSNP)
	    continue;

	 maf = (double)counts[1]/(counts[0]+counts[1]);

	 if(nmiss > total_n*miss || maf<q || 1.0-maf<q)
	    continue;

	 N++;
	 mean /= n_;
	 double variance = mean_sqr/n_ - mean*mean;

	 for(int i=0;i<genotype.Length();i++)
	 {
	    if(genotype[i]==_NAN_)
	       genotype[i]=0.0;
	    else
	       genotype[i] -= mean;
	 }

	 for(int r=0;r<total_n;r++)
	 {
	    for(int c=r;c<total_n;c++)
	    {
	       allPairsX[r][c] += genotype[r]*genotype[c]/variance;
	    }
	 }
      }
   }

   for(int i=0;i<total_n;i++)
      for(int j=i+1;j<total_n;j++)
	 allPairsX[j][i] = allPairsX[i][j];

   if(N==0)
      error("ERROR! No qualified marker was included in calculating empirical kinsihp matrix.\n");
   allPairsX.Multiply(1.0/N);

   printf("completed\n ");
   fprintf(log,"completed\n ");

   printf("  %d markers were included to calculate empirical kinship of X chromosome.\n",N);
   fprintf(log,"  %d markers were included to calculate empirical kinship of X chromosome.\n",N);

   if(OutputKin::outputKin)
   {
      IFILE file = NULL;
      String filename;
      if(PreMeta::outputFile != "")
      {
	 if(PreMeta::outputFile.FastFindLastChar('.')!=-1 || PreMeta::outputFile.FastFindLastChar('/')!=-1)
	    filename.printf("%sEmpirical.KinshipX.gz",(const char *) PreMeta::outputFile);
	 else
	    filename.printf("%s.Empirical.KinshipX.gz",(const char *) PreMeta::outputFile);
      }
      else
	 filename.printf("Empirical.KinshipX.gz");
      file = ifopen(filename, "wt",InputFile::GZIP);
      for (int i = 0; i < total_n; i++)
      {
	 const char * sample = header.getSampleName(genotypedSampleVCF[i]);
	 ifprintf(file,"%s ",sample);
      }
      ifprintf(file,"\n");
      for(int r=0;r<allPairsX.rows;r++)
      {
	 for(int c=0;c<allPairsX.cols;c++)
	 {
	    ifprintf(file,"%g ",allPairsX[r][c]);
	 }
	 ifprintf(file,"\n");
      }
      ifclose(file);
      printf("  Empirical kinshipX saved in %s.\n",filename.c_str());
      fprintf(log,"  Empirical kinshipX saved in %s.\n",filename.c_str());
   }
   printf("\n");
   fprintf(log,"\n");
}
